# Copyright (c) 2018-2022 Trim21 <i@trim21.me>
# Copyright (c) 2008-2014 Erik Svensson <erik.public@gmail.com>
# Licensed under the MIT license.
import enum
import logging
from typing import Dict, Optional, NamedTuple

LOGGER = logging.getLogger("transmission-rpc")
LOGGER.setLevel(logging.ERROR)


def mirror_dict(source: dict) -> dict:
    """
    Creates a dictionary with all values as keys and all keys as values.
    """
    source.update({value: key for key, value in source.items()})
    return source


DEFAULT_TIMEOUT = 30.0

TR_PRI_LOW = -1
TR_PRI_NORMAL = 0
TR_PRI_HIGH = 1

PRIORITY = mirror_dict({"low": TR_PRI_LOW, "normal": TR_PRI_NORMAL, "high": TR_PRI_HIGH})

TR_RATIOLIMIT_GLOBAL = 0  # follow the global settings
TR_RATIOLIMIT_SINGLE = 1  # override the global settings, seeding until a certain ratio
TR_RATIOLIMIT_UNLIMITED = 2  # override the global settings, seeding regardless of ratio

RATIO_LIMIT = mirror_dict(
    {
        "global": TR_RATIOLIMIT_GLOBAL,
        "single": TR_RATIOLIMIT_SINGLE,
        "unlimited": TR_RATIOLIMIT_UNLIMITED,
    }
)

TR_IDLELIMIT_GLOBAL = 0  # follow the global settings
TR_IDLELIMIT_SINGLE = 1  # override the global settings, seeding until a certain idle time
TR_IDLELIMIT_UNLIMITED = 2  # override the global settings, seeding regardless of activity

IDLE_LIMIT = mirror_dict(
    {
        "global": TR_RATIOLIMIT_GLOBAL,
        "single": TR_RATIOLIMIT_SINGLE,
        "unlimited": TR_RATIOLIMIT_UNLIMITED,
    }
)


class Args(NamedTuple):
    type: str
    added_version: int
    removed_version: Optional[int] = None
    previous_argument_name: Optional[str] = None
    next_argument_name: Optional[str] = None
    description: str = ""

    def __repr__(self) -> str:
        return (
            f"Args({self.type!r},"
            f" {self.added_version!r},"
            f" {self.removed_version!r},"
            f" {self.previous_argument_name!r},"
            f" {self.next_argument_name!r},"
            f" {self.description!r})"
        )

    def __str__(self) -> str:
        return f"Args<type={self.type}, {self.added_version}, description={self.description!r})"


class Type:
    number = "number"
    string = "string"
    array = "array"
    boolean = "boolean"
    double = "double"
    object = "object"


trackerListDescription = "A Iterable[Iterable[str]] for a set of announce URLs, each inner list is for a tier"

TORRENT_GET_ARGS: Dict[str, Args] = {
    "activityDate": Args(Type.number, 1, description="Last time of upload or download activity."),
    "addedDate": Args(Type.number, 1, description="The date when this torrent was first added."),
    "bandwidthPriority": Args(Type.number, 5, description="Bandwidth priority. Low (-1), Normal (0) or High (1)."),
    "comment": Args(Type.string, 1, description="Torrent comment."),
    "corruptEver": Args(Type.number, 1, description="Number of bytes of corrupt data downloaded."),
    "creator": Args(Type.string, 1, description="Torrent creator."),
    "dateCreated": Args(Type.number, 1, description="Torrent creation date."),
    "desiredAvailable": Args(Type.number, 1, description="Number of bytes available and left to be downloaded."),
    "doneDate": Args(Type.number, 1, description="The date when the torrent finished downloading."),
    "downloadDir": Args(Type.string, 4, description="The directory path where the torrent is downloaded to."),
    "downloadedEver": Args(Type.number, 1, description="Number of bytes of good data downloaded."),
    "downloadLimit": Args(Type.number, 1, None, None, None, "Download limit in Kbps."),
    "downloadLimitMode": Args(
        Type.number, 1, 5, description="Download limit mode. 0 means global, 1 means single, 2 unlimited."
    ),
    "downloadLimited": Args(Type.boolean, 5, None, None, None, "Download limit is enabled"),
    "editDate": Args(Type.number, 16),
    "error": Args(
        Type.number,
        1,
        description="Kind of error. 0 means OK, 1 means tracker warning, 2 means tracker error, 3 means local error.",
    ),
    "errorString": Args(Type.number, 1, None, None, None, "Error message."),
    "eta": Args(
        Type.number,
        1,
        description="Estimated number of seconds left when downloading or seeding. -1 means not available and -2 means unknown.",
    ),
    "etaIdle": Args(
        Type.number,
        15,
        description="Estimated number of seconds left until the idle time limit is reached. -1 means not available and -2 means unknown.",
    ),
    "files": Args(Type.array, 1, description="Array of file object containing key, bytesCompleted, length and name."),
    "fileStats": Args(
        Type.array, 5, description="Away of file statistics containing bytesCompleted, wanted and priority."
    ),
    "group": Args(Type.string, 17, description="The name of this torrent's bandwidth group"),
    "hashString": Args(Type.string, 1, description="Hashstring unique for the torrent even between sessions."),
    "haveUnchecked": Args(Type.number, 1, None, None, None, "Number of bytes of partial pieces."),
    "haveValid": Args(Type.number, 1, description="Number of bytes of checksum verified data."),
    "honorsSessionLimits": Args(Type.boolean, 5, description="True if session upload limits are honored"),
    "id": Args(Type.number, 1, None, None, None, "Session unique torrent id."),
    "isFinished": Args(Type.boolean, 9, description="True if the torrent is finished. Downloaded and seeded."),
    "isPrivate": Args(Type.boolean, 1, None, None, None, "True if the torrent is private."),
    "isStalled": Args(Type.boolean, 14, description="True if the torrent has stalled (been idle for a long time)."),
    "labels": Args(Type.array, 16, None, None, None, "array of string labels"),
    "leftUntilDone": Args(Type.number, 1, description="Number of bytes left until the download is done."),
    "magnetLink": Args(Type.string, 7, None, None, None, "The magnet link for this torrent."),
    "manualAnnounceTime": Args(Type.number, 1, description="The time until you manually ask for more peers."),
    "maxConnectedPeers": Args(Type.number, 1, None, None, None, "Maximum of connected peers."),
    "metadataPercentComplete": Args(Type.number, 7, description="Download progress of metadata. 0.0 to 1.0."),
    "name": Args(Type.string, 1, None, None, None, "Torrent name."),
    "peer-limit": Args(Type.number, 5, None, None, None, "Maximum number of peers."),
    "peers": Args(Type.array, 2, None, None, None, "Array of peer objects."),
    "peersConnected": Args(Type.number, 1, None, None, None, "Number of peers we are connected to."),
    "peersFrom": Args(Type.object, 1, description="Object containing download peers counts for different peer types."),
    "peersGettingFromUs": Args(Type.number, 1, description="Number of peers we are sending data to."),
    "peersSendingToUs": Args(Type.number, 1, None, None, None, "Number of peers sending to us"),
    "percentComplete": Args(Type.double, 17),
    "percentDone": Args(Type.double, 5, description="Download progress of selected files. 0.0 to 1.0."),
    "pieces": Args(Type.string, 5, description="String with base64 encoded bitfield indicating finished pieces."),
    "pieceCount": Args(Type.number, 1, None, None, None, "Number of pieces."),
    "pieceSize": Args(Type.number, 1, None, None, None, "Number of bytes in a piece."),
    "priorities": Args(Type.array, 1, None, None, None, "Array of file priorities."),
    "primary-mime-type": Args(Type.string, 17),
    "queuePosition": Args(Type.number, 14, None, None, None, "The queue position."),
    "rateDownload": Args(Type.number, 1, None, None, None, "(B/s)"),
    "rateUpload": Args(Type.number, 1, None, None, None, "(B/s)"),
    "recheckProgress": Args(Type.double, 1, None, None, None, "Progress of recheck. 0.0 to 1.0."),
    "secondsDownloading": Args(Type.number, 15, None, None, None, ""),
    "secondsSeeding": Args(Type.number, 15, None, None, None, ""),
    "seedIdleLimit": Args(Type.number, 10, None, None, None, "Idle limit in minutes."),
    "seedIdleMode": Args(Type.number, 10, description="Use global (0), torrent (1), or unlimited (2) limit."),
    "seedRatioLimit": Args(Type.double, 5, None, None, None, "Seed ratio limit."),
    "seedRatioMode": Args(Type.number, 5, description="Use global (0), torrent (1), or unlimited (2) limit."),
    "sizeWhenDone": Args(Type.number, 1, description="Size of the torrent download in bytes."),
    "startDate": Args(Type.number, 1, description="The date when the torrent was last started."),
    "status": Args(Type.number, 1, None, None, None, "Current status, see source"),
    "trackers": Args(Type.array, 1, None, None, None, "Array of tracker objects."),
    "trackerStats": Args(Type.object, 7, description="Array of object containing tracker statistics."),
    "totalSize": Args(Type.number, 1, None, None, None, "Total size of the torrent in bytes"),
    "torrentFile": Args(Type.string, 5, None, None, None, "Path to .torrent file."),
    "uploadedEver": Args(Type.number, 1, None, None, None, "Number of bytes uploaded, ever."),
    "uploadLimit": Args(Type.number, 1, None, None, None, "Upload limit in Kbps"),
    "uploadLimited": Args(Type.boolean, 5, None, None, None, "Upload limit enabled."),
    "uploadRatio": Args(Type.double, 1, None, None, None, "Seed ratio."),
    "wanted": Args(Type.array, 1, description="Array of booleans indicated wanted files."),
    "webseeds": Args(Type.array, 1, None, None, None, "Array of webseeds objects"),
    "webseedsSendingToUs": Args(Type.number, 1, None, None, None, "Number of webseeds seeding to us."),
    "file-count": Args(Type.number, 17),
    "trackerList": Args(Type.array, 17, description=trackerListDescription),
}


class RpcMethod(str, enum.Enum):
    SessionSet = "session-set"
    SessionGet = "session-get"
    SessionStats = "session-stats"

    TorrentGet = "torrent-get"
    TorrentAdd = "torrent-add"
    TorrentSet = "torrent-set"
    TorrentRemove = "torrent-remove"

    TorrentStart = "torrent-start"
    TorrentStartNow = "torrent-start-now"
    TorrentStop = "torrent-stop"

    TorrentVerify = "torrent-verify"
    TorrentReannounce = "torrent-reannounce"

    TorrentSetLocation = "torrent-set-location"
    TorrentRenamePath = "torrent-rename-path"

    QueueMoveTop = "queue-move-top"
    QueueMoveBottom = "queue-move-bottom"
    QueueMoveUp = "queue-move-up"
    QueueMoveDown = "queue-move-down"

    GroupSet = "group-set"
    GroupGet = "group-get"

    FreeSpace = "free-space"

    PortTest = "port-test"

    BlocklistUpdate = "blocklist-update"



import os
import json
import time
import types
import string
import logging
import pathlib
import urllib.parse
from typing import Any, Dict, List, Type, Tuple, Union, Literal, BinaryIO, Iterable, Optional
from urllib.parse import quote

import requests
import requests.auth
import requests.exceptions

from transmission_rpc.error import (
    TransmissionError,
    TransmissionAuthError,
    TransmissionConnectError,
    TransmissionTimeoutError,
)
from transmission_rpc.types import Group, _Timeout
from transmission_rpc.utils import _try_read_torrent, get_torrent_arguments
from transmission_rpc.session import Session, SessionStats
from transmission_rpc.torrent import Torrent
from transmission_rpc.constants import LOGGER, DEFAULT_TIMEOUT, RpcMethod

valid_hash_char = string.digits + string.ascii_letters

_TorrentID = Union[int, str]
_TorrentIDs = Union[_TorrentID, List[_TorrentID], None]


def ensure_location_str(s: Union[str, pathlib.Path]) -> str:
    if isinstance(s, pathlib.Path):
        if s.is_absolute():
            return str(s)

        raise ValueError(
            "using relative `pathlib.Path` as remote path is not supported in v4.",
        )

    return str(s)


def _parse_torrent_id(raw_torrent_id: Union[int, str]) -> Union[int, str]:
    if isinstance(raw_torrent_id, int):
        if raw_torrent_id >= 0:
            return raw_torrent_id
    elif isinstance(raw_torrent_id, str):
        if len(raw_torrent_id) != 40 or (set(raw_torrent_id) - set(valid_hash_char)):
            raise ValueError(f"torrent ids {raw_torrent_id} is not valid torrent id")
        return raw_torrent_id
    raise ValueError(f"{raw_torrent_id} is not valid torrent id")


def _parse_torrent_ids(args: Any) -> Union[str, List[Union[str, int]]]:
    if args is None:
        return []
    if isinstance(args, int):
        return [_parse_torrent_id(args)]
    if isinstance(args, str):
        if args == "recently-active":
            return args
        return [_parse_torrent_id(args)]
    if isinstance(args, (list, tuple)):
        return [_parse_torrent_id(item) for item in args]
    raise ValueError(f"Invalid torrent id {args}")


class Client:
    def __init__(
        self,
        *,
        protocol: Literal["http", "https"] = "http",
        username: Optional[str] = None,
        password: Optional[str] = None,
        host: str = "127.0.0.1",
        port: int = 9091,
        path: str = "/transmission/rpc",
        timeout: Union[int, float] = DEFAULT_TIMEOUT,
        logger: logging.Logger = LOGGER,
    ):
        if isinstance(logger, logging.Logger):
            self.logger = logger
        else:
            raise TypeError(
                "logger must be instance of `logging.Logger`, default: logging.getLogger('transmission-rpc')"
            )
        self._query_timeout: _Timeout = timeout

        username = quote(username or "", safe="$-_.+!*'(),;&=", encoding="utf8") if username else ""
        password = ":" + quote(password or "", safe="$-_.+!*'(),;&=", encoding="utf8") if password else ""
        auth = f"{username}{password}@" if (username or password) else ""

        if path == "/transmission/":
            path = "/transmission/rpc"

        url = urllib.parse.urlunparse((protocol, f"{auth}{host}:{port}", path, None, None, None))
        self.url = str(url)
        self._sequence = 0
        self.raw_session: Dict[str, Any] = {}
        self.session_id = "0"
        self.server_version: Optional[Tuple[int, int, Optional[str]]] = None
        self.protocol_version: int = 17  # default 17
        self._http_session = requests.Session()
        self._http_session.trust_env = False
        self.get_session()
        self.torrent_get_arguments = get_torrent_arguments(self.rpc_version)

    @property
    def timeout(self) -> _Timeout:
        """
        Get current timeout for HTTP queries.
        """
        return self._query_timeout

    @timeout.setter
    def timeout(self, value: _Timeout) -> None:
        """
        Set timeout for HTTP queries.
        """
        if isinstance(value, (tuple, list)):
            if len(value) != 2:
                raise ValueError("timeout tuple can only include 2 numbers elements")
            for v in value:
                if not isinstance(v, (float, int)):
                    raise ValueError("element of timeout tuple can only be int of float")
            self._query_timeout = (value[0], value[1])  # for type checker
        elif value is None:
            self._query_timeout = DEFAULT_TIMEOUT
        else:
            self._query_timeout = float(value)

    @timeout.deleter
    def timeout(self) -> None:
        """
        Reset the HTTP query timeout to the default.
        """
        self._query_timeout = DEFAULT_TIMEOUT

    @property
    def _http_header(self) -> Dict[str, str]:
        return {"x-transmission-session-id": self.session_id}

    def _http_query(self, query: dict, timeout: Optional[_Timeout] = None) -> str:
        """
        Query Transmission through HTTP.
        """
        request_count = 0
        if timeout is None:
            timeout = self.timeout
        while True:
            if request_count >= 10:
                raise TransmissionError("too much request, try enable logger to see what happened")
            self.logger.debug(
                {
                    "url": self.url,
                    "headers": self._http_header,
                    "data": query,
                    "timeout": timeout,
                }
            )
            request_count += 1
            try:
                r = self._http_session.post(
                    self.url,
                    headers=self._http_header,
                    json=query,
                    timeout=timeout,
                )
            except requests.exceptions.Timeout as e:
                raise TransmissionTimeoutError("timeout when connection to transmission daemon") from e
            except requests.exceptions.ConnectionError as e:
                raise TransmissionConnectError(f"can't connect to transmission daemon: {str(e)}") from e

            self.session_id = r.headers.get("X-Transmission-Session-Id", "0")
            self.logger.debug(r.text)
            if r.status_code in {401, 403}:
                self.logger.debug(r.request.headers)
                raise TransmissionAuthError("transmission daemon require auth", original=r)
            if r.status_code != 409:
                return r.text

    def _request(
        self,
        method: RpcMethod,
        arguments: Optional[Dict[str, Any]] = None,
        ids: Optional[_TorrentIDs] = None,
        require_ids: bool = False,
        timeout: Optional[_Timeout] = None,
    ) -> dict:
        """
        Send json-rpc request to Transmission using http POST
        """
        if not isinstance(method, str):
            raise ValueError("request takes method as string")
        if arguments is None:
            arguments = {}
        if not isinstance(arguments, dict):
            raise ValueError("request takes arguments as dict")

        ids = _parse_torrent_ids(ids)
        if len(ids) > 0:
            arguments["ids"] = ids
        elif require_ids:
            raise ValueError("request require ids")

        query = {"tag": self._sequence, "method": method, "arguments": arguments}

        self._sequence += 1
        start = time.time()
        http_data = self._http_query(query, timeout)
        elapsed = time.time() - start
        self.logger.info("http request took %.3f s", elapsed)

        try:
            data: dict = json.loads(http_data)
        except ValueError as error:
            self.logger.error("Error: %s", str(error))
            self.logger.error('Request: "%s"', query)
            self.logger.error('HTTP data: "%s"', http_data)
            raise ValueError from error

        self.logger.debug(json.dumps(data, indent=2))
        if "result" not in data:
            raise TransmissionError("Query failed without result.")

        if data["result"] != "success":
            raise TransmissionError(f'Query failed with result "{data["result"]}".')
        results = {}
        if method == RpcMethod.TorrentGet:
            return data["arguments"]
        elif method == RpcMethod.TorrentAdd:
            item = None
            if "torrent-added" in data["arguments"]:
                item = data["arguments"]["torrent-added"]
            elif "torrent-duplicate" in data["arguments"]:
                item = data["arguments"]["torrent-duplicate"]
            if item:
                results[item["id"]] = Torrent(fields=item)
            else:
                raise TransmissionError("Invalid torrent-add response.")
        elif method == RpcMethod.SessionGet:
            self.raw_session.update(data["arguments"])
        elif method == RpcMethod.SessionStats:
            # older versions of T has the return data in "session-stats"
            if "session-stats" in data["arguments"]:
                return data["arguments"]["session-stats"]
            else:
                return data["arguments"]
        elif method in (
            RpcMethod.PortTest,
            RpcMethod.BlocklistUpdate,
            RpcMethod.FreeSpace,
            RpcMethod.TorrentRenamePath,
        ):
            results = data["arguments"]
        else:
            return data

        return results

    def _update_server_version(self) -> None:
        """Decode the Transmission version string, if available."""
        self.protocol_version = self.raw_session["rpc-version"]

    @property
    def rpc_version(self) -> int:
        """
        Get the Transmission RPC version. Trying to deduct if the server don't have a version value.
        """
        return self.protocol_version

    def _rpc_version_warning(self, required_version: int) -> None:
        """
        Add a warning to the log if the Transmission RPC version is lower then the provided version.
        """
        if self.rpc_version < required_version:
            self.logger.warning(
                "Using feature not supported by server. RPC version for server %d, feature introduced in %d.",
                self.rpc_version,
                required_version,
            )

    def add_torrent(
        self,
        torrent: Union[BinaryIO, str, bytes, pathlib.Path],
        timeout: Optional[_Timeout] = None,
        *,
        download_dir: Optional[str] = None,
        files_unwanted: Optional[List[int]] = None,
        files_wanted: Optional[List[int]] = None,
        paused: Optional[bool] = None,
        peer_limit: Optional[int] = None,
        priority_high: Optional[List[int]] = None,
        priority_low: Optional[List[int]] = None,
        priority_normal: Optional[List[int]] = None,
        cookies: Optional[str] = None,
        labels: Optional[Iterable[str]] = None,
        bandwidthPriority: Optional[int] = None,
    ) -> Torrent:
        """
        Add torrent to transfers list. ``torrent`` can be:

        - ``http://``, ``https://`` or  ``magnet:`` URL
        - torrent file-like object in binary mode
        - bytes of torrent content
        - ``pathlib.Path`` for local torrent file, will be read and encoded as base64.

        Warnings
        --------
        base64 string or ``file://`` protocol URL are not supported in v4.

        Parameters
        ----------
        torrent:
            torrent to add
        timeout:
            request timeout
        labels:
            Array of string labels.
            Add in rpc 17.
        bandwidthPriority:
            Priority for this transfer.
        cookies:
            One or more HTTP cookie(s).
        download_dir:
            The directory where the downloaded contents will be saved in.
        files_unwanted:
            A list of file id's that shouldn't be downloaded.
        files_wanted:
            A list of file id's that should be downloaded.
        paused:
            If True, does not start the transfer when added.
            Magnet url will always start to downloading torrents.
        peer_limit:
            Maximum number of peers allowed.
        priority_high:
            A list of file id's that should have high priority.
        priority_low:
            A list of file id's that should have low priority.
        priority_normal:
            A list of file id's that should have normal priority.
        """
        if torrent is None:
            raise ValueError("add_torrent requires data or a URI.")

        kwargs: Dict[str, Any] = {}
        if download_dir is not None:
            kwargs["download-dir"] = download_dir

        if files_unwanted is not None:
            kwargs["files-unwanted"] = files_unwanted

        if files_wanted is not None:
            kwargs["files-wanted"] = files_wanted

        if paused is not None:
            kwargs["paused"] = paused

        if peer_limit is not None:
            kwargs["peer-limit"] = peer_limit

        if priority_high is not None:
            kwargs["priority-high"] = priority_high

        if priority_low is not None:
            kwargs["priority-low"] = priority_low

        if priority_normal is not None:
            kwargs["priority-normal"] = priority_normal

        if bandwidthPriority is not None:
            kwargs["bandwidthPriority"] = bandwidthPriority

        if cookies is not None:
            kwargs["cookies"] = cookies

        if labels is not None:
            self._rpc_version_warning(17)
            kwargs["labels"] = list(labels)

        if torrent_data := _try_read_torrent(torrent):
            kwargs["metainfo"] = torrent_data
        else:
            kwargs["filename"] = torrent

        return list(self._request(RpcMethod.TorrentAdd, kwargs, timeout=timeout).values())[0]

    def remove_torrent(self, ids: _TorrentIDs, delete_data: bool = False, timeout: Optional[_Timeout] = None) -> None:
        """
        remove torrent(s) with provided id(s). Local data is removed if
        delete_data is True, otherwise not.
        """
        self._request(
            RpcMethod.TorrentRemove,
            {"delete-local-data": delete_data},
            ids,
            True,
            timeout=timeout,
        )

    def start_torrent(self, ids: _TorrentIDs, bypass_queue: bool = False, timeout: Optional[_Timeout] = None) -> None:
        """Start torrent(s) with provided id(s)"""
        method = RpcMethod.TorrentStart
        if bypass_queue:
            method = RpcMethod.TorrentStartNow
        self._request(method, {}, ids, True, timeout=timeout)

    def start_all(self, bypass_queue: bool = False, timeout: Optional[_Timeout] = None) -> None:
        """Start all torrents respecting the queue order"""
        method = RpcMethod.TorrentStart
        if bypass_queue:
            method = RpcMethod.TorrentStartNow
        torrent_list = sorted(self.get_torrents(), key=lambda t: t.queue_position)
        self._request(
            method,
            {},
            ids=[x.id for x in torrent_list],
            require_ids=True,
            timeout=timeout,
        )

    def stop_torrent(self, ids: _TorrentIDs, timeout: Optional[_Timeout] = None) -> None:
        """stop torrent(s) with provided id(s)"""
        self._request(RpcMethod.TorrentStop, {}, ids, True, timeout=timeout)

    def verify_torrent(self, ids: _TorrentIDs, timeout: Optional[_Timeout] = None) -> None:
        """verify torrent(s) with provided id(s)"""
        self._request(RpcMethod.TorrentVerify, {}, ids, True, timeout=timeout)

    def reannounce_torrent(self, ids: _TorrentIDs, timeout: Optional[_Timeout] = None) -> None:
        """Reannounce torrent(s) with provided id(s)"""
        self._request(RpcMethod.TorrentReannounce, {}, ids, True, timeout=timeout)

    def get_torrent(
        self,
        torrent_id: _TorrentID,
        arguments: Optional[Iterable[str]] = None,
        timeout: Optional[_Timeout] = None,
    ) -> Torrent:
        """
        Get information for torrent with provided id.
        ``arguments`` contains a list of field names to be returned, when None
        all fields are requested. See the Torrent class for more information.

        new argument ``format`` in rpc_version 16 is unnecessarily
        and this lib can't handle table response, So it's unsupported.

        Returns a Torrent object with the requested fields.


        Note
        ----
        It's recommended that you use torrent's info_hash as torrent id. torrent's info_hash will never change.

        Parameters
        ----------
        torrent_id:
            torrent id can be an int or a torrent info_hash (hash_string of torrent object).

        arguments:
            fetched torrent arguments, in most cases you don't need to set this,
            transmission-rpc will fetch all torrent fields it supported.

        timeout:
            requests timeout
        """
        if arguments:
            arguments = list(set(arguments) | {"id", "hashString"})
        else:
            arguments = self.torrent_get_arguments
        torrent_id = _parse_torrent_id(torrent_id)
        if torrent_id is None:
            raise ValueError("Invalid id")

        result = self._request(
            RpcMethod.TorrentGet,
            {"fields": arguments},
            torrent_id,
            require_ids=True,
            timeout=timeout,
        )

        for torrent in result["torrents"]:
            if torrent.get("hashString") == torrent_id or torrent.get("id") == torrent_id:
                return Torrent(fields=torrent)
        raise KeyError("Torrent not found in result")

    def get_torrents(
        self,
        ids: Optional[_TorrentIDs] = None,
        arguments: Optional[Iterable[str]] = None,
        timeout: Optional[_Timeout] = None,
    ) -> List[Torrent]:
        """
        Get information for torrents with provided ids. For more information see ``get_torrent``.

        Returns a list of Torrent object.
        """
        if arguments:
            arguments = list(set(arguments) | {"id", "hashString"})
        else:
            arguments = self.torrent_get_arguments
        return [
            Torrent(fields=x)
            for x in self._request(RpcMethod.TorrentGet, {"fields": arguments}, ids, timeout=timeout)["torrents"]
        ]

    def get_recently_active_torrents(
        self, arguments: Optional[Iterable[str]] = None, timeout: Optional[_Timeout] = None
    ) -> Tuple[List[Torrent], List[int]]:
        """
        Get information for torrents for recently active torrent. If you want to get recently-removed
        torrents. you should use this method.

        Returns
        -------
        active_torrents: List[Torrent]
            List of recently active torrents
        removed_torrents: List[int]
            List of torrent-id of recently-removed torrents.
        """
        if arguments:
            arguments = list(set(arguments) | {"id", "hashString"})
        else:
            arguments = self.torrent_get_arguments

        result = self._request(RpcMethod.TorrentGet, {"fields": arguments}, "recently-active", timeout=timeout)

        return [Torrent(fields=x) for x in result["torrents"]], result["removed"]

    def change_torrent(
        self,
        ids: _TorrentIDs,
        timeout: Optional[_Timeout] = None,
        *,
        bandwidth_priority: Optional[int] = None,
        download_limit: Optional[int] = None,
        download_limited: Optional[bool] = None,
        upload_limit: Optional[int] = None,
        upload_limited: Optional[bool] = None,
        files_unwanted: Optional[Iterable[int]] = None,
        files_wanted: Optional[Iterable[int]] = None,
        honors_session_limits: Optional[bool] = None,
        location: Optional[str] = None,
        peer_limit: Optional[int] = None,
        priority_high: Optional[Iterable[int]] = None,
        priority_low: Optional[Iterable[int]] = None,
        priority_normal: Optional[Iterable[int]] = None,
        queue_position: Optional[int] = None,
        seed_idle_limit: Optional[int] = None,
        seed_idle_mode: Optional[int] = None,
        seed_ratio_limit: Optional[float] = None,
        seed_ratio_mode: Optional[int] = None,
        tracker_add: Optional[Iterable[str]] = None,
        tracker_remove: Optional[Iterable[int]] = None,
        tracker_replace: Optional[Iterable[Tuple[int, str]]] = None,
        labels: Optional[Iterable[str]] = None,
        group: Optional[str] = None,
        tracker_list: Optional[Iterable[Iterable[str]]] = None,
        **kwargs: Any,
    ) -> None:
        """Change torrent parameters for the torrent(s) with the supplied id's.

        Parameters
        ----------
        ids
            torrent(s) to change.
        timeout
            requesst timeout.
        honors_session_limits
            true if session upload limits are honored.
        location
            new location of the torrent's content
        peer_limit
            maximum number of peers
        queue_position
            position of this torrent in its queue [0...n)
        files_wanted
            Array of file id to download.
        files_unwanted
            Array of file id to not download.
        download_limit
            maximum download speed (KBps)
        download_limited
            true if ``download_limit`` is honored
        upload_limit
            maximum upload speed (KBps)
        upload_limited
            true if ``upload_limit`` is honored
        bandwidth_priority
            Priority for this transfer.
        priority_high
            list of file id to set high download priority
        priority_low
            list of file id to set low download priority
        priority_normal
            list of file id to set normal download priority
        seed_ratio_limit
            Seed inactivity limit in minutes.
        seed_ratio_mode
            Which ratio to use.

            0 = Use session limit

            1 = Use transfer limit

            2 = Disable limit.
        seed_idle_limit
            torrent-level seeding ratio
        seed_idle_mode
            Seed inactivity mode.

            0 = Use session limit

            1 = Use transfer limit

            2 = Disable limit.
        tracker_add
            Array of string with announce URLs to add.
        tracker_remove
            Array of ids of trackers to remove.
        tracker_replace
            Array of (id, url) tuples where the announce URL should be replaced.
        labels
            Array of string labels.
            Add in rpc 16.
        group
            The name of this torrent's bandwidth group.
            Add in rpc 17.
        tracker_list
            A ``Iterable[Iterable[str]]``, each ``Iterable[str]`` for a tracker tier.
            Add in rpc 17.


        Warnings
        ----
        ``kwargs`` is for the future features not supported yet, it's not compatibility promising.

        it will be bypassed to request arguments.
        """

        args: Dict[str, Any] = {}

        if bandwidth_priority is not None:
            args["bandwidthPriority"] = bandwidth_priority

        if download_limit is not None:
            args["downloadLimit"] = download_limit
        if download_limited is not None:
            args["downloadLimited"] = download_limited
        if upload_limit is not None:
            args["uploadLimit"] = upload_limit
        if upload_limited is not None:
            args["uploadLimited"] = upload_limited
        if files_unwanted is not None:
            args["files-unwanted"] = list(files_unwanted)
        if files_wanted is not None:
            args["files-wanted"] = list(files_wanted)
        if honors_session_limits is not None:
            args["honorsSessionLimits"] = honors_session_limits
        if location is not None:
            args["location"] = location
        if peer_limit is not None:
            args["peer-limit"] = peer_limit
        if priority_high is not None:
            args["priority-high"] = list(priority_high)
        if priority_low is not None:
            args["priority-low"] = list(priority_low)
        if priority_normal is not None:
            args["priority-normal"] = list(priority_normal)
        if queue_position is not None:
            args["queuePosition"] = queue_position
        if seed_idle_limit is not None:
            args["seedIdleLimit"] = seed_idle_limit
        if seed_idle_mode is not None:
            args["seedIdleMode"] = seed_idle_mode
        if seed_ratio_limit is not None:
            args["seedRatioLimit"] = seed_ratio_limit
        if seed_ratio_mode is not None:
            args["seedRatioMode"] = seed_ratio_mode
        if tracker_add is not None:
            args["trackerAdd"] = tracker_add
        if tracker_remove is not None:
            args["trackerRemove"] = tracker_remove
        if tracker_replace is not None:
            args["trackerReplace"] = tracker_replace
        if labels is not None:
            self._rpc_version_warning(16)
            args["labels"] = list(labels)

        if tracker_list is not None:
            self._rpc_version_warning(17)
            args["trackerList"] = " ".join("\n".join(x) for x in tracker_list)

        if group is not None:
            self._rpc_version_warning(17)
            args["group"] = str(group)

        args.update(kwargs)

        if args:
            self._request(RpcMethod.TorrentSet, args, ids, True, timeout=timeout)
        else:
            ValueError("No arguments to set")

    def move_torrent_data(
        self,
        ids: _TorrentIDs,
        location: Union[str, pathlib.Path],
        timeout: Optional[_Timeout] = None,
    ) -> None:
        """Move torrent data to the new location."""
        args = {"location": ensure_location_str(location), "move": True}
        self._request(RpcMethod.TorrentSetLocation, args, ids, True, timeout=timeout)

    def locate_torrent_data(
        self,
        ids: _TorrentIDs,
        location: Union[str, pathlib.Path],
        timeout: Optional[_Timeout] = None,
    ) -> None:
        """Locate torrent data at the provided location."""
        args = {"location": ensure_location_str(location), "move": False}
        self._request(RpcMethod.TorrentSetLocation, args, ids, True, timeout=timeout)

    def rename_torrent_path(
        self,
        torrent_id: _TorrentID,
        location: Union[str, pathlib.Path],
        name: str,
        timeout: Optional[_Timeout] = None,
    ) -> Tuple[str, str]:
        """
        https://github.com/transmission/transmission/blob/main/docs/rpc-spec.md#37-renaming-a-torrents-path

        This method can only be called on single torrent.
        """
        self._rpc_version_warning(15)
        torrent_id = _parse_torrent_id(torrent_id)
        name = name.strip()  # https://github.com/trim21/transmission-rpc/issues/185
        dirname = os.path.dirname(name)
        if len(dirname) > 0:
            raise ValueError("Target name cannot contain a path delimiter")
        result = self._request(
            RpcMethod.TorrentRenamePath,
            {"path": ensure_location_str(location), "name": name},
            torrent_id,
            True,
            timeout=timeout,
        )
        return result["path"], result["name"]

    def queue_top(self, ids: _TorrentIDs, timeout: Optional[_Timeout] = None) -> None:
        """Move transfer to the top of the queue:_Timeout."""
        self._request(RpcMethod.QueueMoveTop, ids=ids, require_ids=True, timeout=timeout)

    def queue_bottom(self, ids: _TorrentIDs, timeout: Optional[_Timeout] = None) -> None:
        """Move transfer to the bottom of the queue."""
        self._request(RpcMethod.QueueMoveBottom, ids=ids, require_ids=True, timeout=timeout)

    def queue_up(self, ids: _TorrentIDs, timeout: Optional[_Timeout] = None) -> None:
        """Move transfer up in the queue."""
        self._request(RpcMethod.QueueMoveUp, ids=ids, require_ids=True, timeout=timeout)

    def queue_down(self, ids: _TorrentIDs, timeout: Optional[_Timeout] = None) -> None:
        """Move transfer down in the queue."""
        self._request(RpcMethod.QueueMoveDown, ids=ids, require_ids=True, timeout=timeout)

    def get_session(self, timeout: Optional[_Timeout] = None) -> Session:
        """
        Get session parameters. See the Session class for more information.
        """
        self._request(RpcMethod.SessionGet, timeout=timeout)
        self._update_server_version()
        return Session(fields=self.raw_session)

    def set_session(
        self,
        timeout: Optional[_Timeout] = None,
        *,
        alt_speed_down: Optional[int] = None,
        alt_speed_enabled: Optional[bool] = None,
        alt_speed_time_begin: Optional[int] = None,
        alt_speed_time_day: Optional[int] = None,
        alt_speed_time_enabled: Optional[bool] = None,
        alt_speed_time_end: Optional[int] = None,
        alt_speed_up: Optional[int] = None,
        blocklist_enabled: Optional[bool] = None,
        blocklist_url: Optional[str] = None,
        cache_size_mb: Optional[int] = None,
        dht_enabled: Optional[bool] = None,
        default_trackers: Optional[Iterable[str]] = None,
        download_dir: Optional[str] = None,
        download_queue_enabled: Optional[bool] = None,
        download_queue_size: Optional[int] = None,
        encryption: Optional[Literal["required", "preferred", "tolerated"]] = None,
        idle_seeding_limit: Optional[int] = None,
        idle_seeding_limit_enabled: Optional[bool] = None,
        incomplete_dir: Optional[str] = None,
        incomplete_dir_enabled: Optional[bool] = None,
        lpd_enabled: Optional[bool] = None,
        peer_limit_global: Optional[int] = None,
        peer_limit_per_torrent: Optional[int] = None,
        peer_port: Optional[int] = None,
        peer_port_random_on_start: Optional[bool] = None,
        pex_enabled: Optional[bool] = None,
        port_forwarding_enabled: Optional[bool] = None,
        queue_stalled_enabled: Optional[bool] = None,
        queue_stalled_minutes: Optional[int] = None,
        rename_partial_files: Optional[bool] = None,
        script_torrent_done_enabled: Optional[bool] = None,
        script_torrent_done_filename: Optional[str] = None,
        seed_queue_enabled: Optional[bool] = None,
        seed_queue_size: Optional[int] = None,
        seed_ratio_limit: Optional[int] = None,
        seed_ratio_limited: Optional[bool] = None,
        speed_limit_down: Optional[int] = None,
        speed_limit_down_enabled: Optional[bool] = None,
        speed_limit_up: Optional[int] = None,
        speed_limit_up_enabled: Optional[bool] = None,
        start_added_torrents: Optional[bool] = None,
        trash_original_torrent_files: Optional[bool] = None,
        utp_enabled: Optional[bool] = None,
        script_torrent_done_seeding_filename: Optional[str] = None,
        script_torrent_done_seeding_enabled: Optional[bool] = None,
        script_torrent_added_enabled: Optional[bool] = None,
        script_torrent_added_filename: Optional[str] = None,
        **kwargs: Any,
    ) -> None:
        """
        Set session parameters.

        Parameters
        ----------
        timeout
            request timeout
        alt_speed_down:
            max global download speed (KBps)
        alt_speed_enabled:
            true means use the alt speeds
        alt_speed_time_begin:
            Time when alternate speeds should be enabled. Minutes after midnight.
        alt_speed_time_day:
            Enables alternate speeds scheduling these days.
        alt_speed_time_enabled:
            Enables alternate speeds scheduling.
        alt_speed_time_end:
            Time when alternate speeds should be disabled. Minutes after midnight.
        alt_speed_up:
            Alternate session upload speed limit (in Kib/s).
        blocklist_enabled:
            Enables the block list
        blocklist_url:
            Location of the block list. Updated with blocklist-update.
        cache_size_mb:
            The maximum size of the disk cache in MB
        default_trackers:
            List of default trackers to use on public torrents.
        dht_enabled:
            Enables DHT.
        download_dir:
            Set the session download directory.
        download_queue_enabled:
            Enables download queue.
        download_queue_size:
            Number of slots in the download queue.
        encryption:
            Set the session encryption mode, one of ``required``, ``preferred`` or ``tolerated``.
        idle_seeding_limit:
            The default seed inactivity limit in minutes.
        idle_seeding_limit_enabled:
            Enables the default seed inactivity limit
        incomplete_dir:
            The path to the directory of incomplete transfer data.
        incomplete_dir_enabled:
            Enables the incomplete transfer data directory,
            Otherwise data for incomplete transfers are stored in the download target.
        lpd_enabled:
            Enables local peer discovery for public torrents.
        peer_limit_global:
            Maximum number of peers.
        peer_limit_per_torrent:
            Maximum number of peers per transfer.
        peer_port:
            Peer port.
        peer_port_random_on_start:
            Enables randomized peer port on start of Transmission.
        pex_enabled:
            Allowing PEX in public torrents.
        port_forwarding_enabled:
            Enables port forwarding.
        queue_stalled_enabled:
            Enable tracking of stalled transfers.
        queue_stalled_minutes:
            Number of minutes of idle that marks a transfer as stalled.
        rename_partial_files:
            Appends ".part" to incomplete files

        seed_queue_enabled:
            Enables upload queue.
        seed_queue_size:
            Number of slots in the upload queue.
        seed_ratio_limit:
            Seed ratio limit. 1.0 means 1:1 download and upload ratio.
        seed_ratio_limited:
            Enables seed ration limit.
        speed_limit_down:
            Download speed limit (in Kib/s).
        speed_limit_down_enabled:
            Enables download speed limiting.
        speed_limit_up:
            Upload speed limit (in Kib/s).
        speed_limit_up_enabled:
            Enables upload speed limiting.
        start_added_torrents:
            Added torrents will be started right away.
        trash_original_torrent_files:
            The .torrent file of added torrents will be deleted.
        utp_enabled:
            Enables Micro Transport Protocol (UTP).
        script_torrent_done_enabled:
            Whether to call the "done" script.
        script_torrent_done_filename:
            Filename of the script to run when the transfer is done.
        script_torrent_added_filename:
            filename of the script to run
        script_torrent_added_enabled:
            whether or not to call the ``added`` script
        script_torrent_done_seeding_enabled:
            whether or not to call the ``seeding-done`` script
        script_torrent_done_seeding_filename:
            filename of the script to run

        Warnings
        ----
        ``kwargs`` is for the future features not supported yet, it's not compatibility promising.

        it will be bypassed to request arguments.
        """
        args: Dict[str, Any] = {}

        if alt_speed_down is not None:
            args["alt-speed-down"] = alt_speed_down
        if alt_speed_enabled is not None:
            args["alt-speed-enabled"] = alt_speed_enabled
        if alt_speed_time_begin is not None:
            args["alt-speed-time-begin"] = alt_speed_time_begin
        if alt_speed_time_day is not None:
            args["alt-speed-time-day"] = alt_speed_time_day
        if alt_speed_time_enabled is not None:
            args["alt-speed-time-enabled"] = alt_speed_time_enabled
        if alt_speed_time_end is not None:
            args["alt-speed-time-end"] = alt_speed_time_end
        if alt_speed_up is not None:
            args["alt-speed-up"] = alt_speed_up
        if blocklist_enabled is not None:
            args["blocklist-enabled"] = blocklist_enabled
        if blocklist_url is not None:
            args["blocklist-url"] = blocklist_url
        if cache_size_mb is not None:
            args["cache-size-mb"] = cache_size_mb
        if dht_enabled is not None:
            args["dht-enabled"] = dht_enabled
        if download_dir is not None:
            args["download-dir"] = download_dir
        if download_queue_enabled is not None:
            args["download-queue-enabled"] = download_queue_enabled
        if download_queue_size is not None:
            args["download-queue-size"] = download_queue_size
        if encryption is not None:
            if encryption not in ["required", "preferred", "tolerated"]:
                raise ValueError("Invalid encryption value")
            args["encryption"] = encryption
        if idle_seeding_limit_enabled is not None:
            args["idle-seeding-limit-enabled"] = idle_seeding_limit_enabled
        if idle_seeding_limit is not None:
            args["idle-seeding-limit"] = idle_seeding_limit
        if incomplete_dir is not None:
            args["incomplete-dir"] = incomplete_dir
        if incomplete_dir_enabled is not None:
            args["incomplete-dir-enabled"] = incomplete_dir_enabled
        if lpd_enabled is not None:
            args["lpd-enabled"] = lpd_enabled
        if peer_limit_global is not None:
            args["peer-limit-global"] = peer_limit_global
        if peer_limit_per_torrent is not None:
            args["peer-limit-per-torrent"] = peer_limit_per_torrent
        if peer_port_random_on_start is not None:
            args["peer-port-random-on-start"] = peer_port_random_on_start
        if peer_port is not None:
            args["peer-port"] = peer_port
        if pex_enabled is not None:
            args["pex-enabled"] = pex_enabled
        if port_forwarding_enabled is not None:
            args["port-forwarding-enabled"] = port_forwarding_enabled
        if queue_stalled_enabled is not None:
            args["queue-stalled-enabled"] = queue_stalled_enabled
        if queue_stalled_minutes is not None:
            args["queue-stalled-minutes"] = queue_stalled_minutes
        if rename_partial_files is not None:
            args["rename-partial-files"] = rename_partial_files
        if script_torrent_done_enabled is not None:
            args["script-torrent-done-enabled"] = script_torrent_done_enabled
        if script_torrent_done_filename is not None:
            args["script-torrent-done-filename"] = script_torrent_done_filename
        if seed_queue_enabled is not None:
            args["seed-queue-enabled"] = seed_queue_enabled
        if seed_queue_size is not None:
            args["seed-queue-size"] = seed_queue_size
        if seed_ratio_limit is not None:
            args["seedRatioLimit"] = seed_ratio_limit
        if seed_ratio_limited is not None:
            args["seedRatioLimited"] = seed_ratio_limited
        if speed_limit_down is not None:
            args["speed-limit-down"] = speed_limit_down
        if speed_limit_down_enabled is not None:
            args["speed-limit-down-enabled"] = speed_limit_down_enabled
        if speed_limit_up is not None:
            args["speed-limit-up"] = speed_limit_up
        if speed_limit_up_enabled is not None:
            args["speed-limit-up-enabled"] = speed_limit_up_enabled
        if start_added_torrents is not None:
            args["start-added-torrents"] = start_added_torrents
        if trash_original_torrent_files is not None:
            args["trash-original-torrent-files"] = trash_original_torrent_files
        if utp_enabled is not None:
            args["utp-enabled"] = utp_enabled

        if default_trackers is not None:
            self._rpc_version_warning(17)
            args["default-trackers"] = list(default_trackers)

        if script_torrent_done_seeding_filename is not None:
            self._rpc_version_warning(17)
            args["script-torrent-done-seeding-filename"] = script_torrent_done_seeding_filename
        if script_torrent_done_seeding_enabled is not None:
            self._rpc_version_warning(17)
            args["script-torrent-done-seeding-enabled"] = script_torrent_done_seeding_enabled
        if script_torrent_added_enabled is not None:
            self._rpc_version_warning(17)
            args["script-torrent-added-enabled"] = script_torrent_added_enabled
        if script_torrent_added_filename is not None:
            self._rpc_version_warning(17)
            args["script-torrent-added-filename"] = script_torrent_added_filename

        args.update(kwargs)

        if args:
            self._request(RpcMethod.SessionSet, args, timeout=timeout)

    def blocklist_update(self, timeout: Optional[_Timeout] = None) -> Optional[int]:
        """Update block list. Returns the size of the block list."""
        result = self._request(RpcMethod.BlocklistUpdate, timeout=timeout)
        return result.get("blocklist-size")

    def port_test(self, timeout: Optional[_Timeout] = None) -> Optional[bool]:
        """
        Tests to see if your incoming peer port is accessible from the
        outside world.
        """
        result = self._request(RpcMethod.PortTest, timeout=timeout)
        return result.get("port-is-open")

    def free_space(self, path: Union[str, pathlib.Path], timeout: Optional[_Timeout] = None) -> Optional[int]:
        """
        Get the amount of free space (in bytes) at the provided location.
        """
        self._rpc_version_warning(15)
        path = ensure_location_str(path)
        result: Dict[str, Any] = self._request(RpcMethod.FreeSpace, {"path": path}, timeout=timeout)
        if result["path"] == path:
            return result["size-bytes"]
        return None

    def session_stats(self, timeout: Optional[_Timeout] = None) -> SessionStats:
        """Get session statistics"""
        result = self._request(RpcMethod.SessionStats, timeout=timeout)
        return SessionStats(fields=result)

    def set_group(
        self,
        name: str,
        *,
        timeout: Optional[_Timeout] = None,
        honors_session_limits: Optional[bool] = None,
        speed_limit_down: Optional[int] = None,
        speed_limit_up_enabled: Optional[bool] = None,
        speed_limit_up: Optional[int] = None,
        speed_limit_down_enabled: Optional[bool] = None,
    ) -> None:
        self._rpc_version_warning(17)
        arguments: Dict[str, Any] = {"name": name}

        if honors_session_limits is not None:
            arguments["honorsSessionLimits"] = honors_session_limits

        if speed_limit_down is not None:
            arguments["speed-limit-down"] = speed_limit_down

        if speed_limit_up_enabled is not None:
            arguments["speed-limit-up-enabled"] = speed_limit_up_enabled

        if speed_limit_up is not None:
            arguments["speed-limit-up"] = speed_limit_up

        if speed_limit_down_enabled is not None:
            arguments["speed-limit-down-enabled"] = speed_limit_down_enabled

        self._request(RpcMethod.GroupSet, arguments, timeout=timeout)

    def get_group(self, name: str, *, timeout: Optional[_Timeout] = None) -> Optional[Group]:
        self._rpc_version_warning(17)
        result: Dict[str, Any] = self._request(RpcMethod.GroupGet, {"group": name}, timeout=timeout)

        if result["arguments"]["group"]:
            return Group(fields=result["arguments"]["group"][0])
        return None

    def get_groups(self, name: Optional[List[str]] = None, *, timeout: Optional[_Timeout] = None) -> Dict[str, Group]:
        payload = {}
        if name is not None:
            payload = {"group": name}

        result: Dict[str, Any] = self._request(RpcMethod.GroupGet, payload, timeout=timeout)

        return {x["name"]: Group(fields=x) for x in result["arguments"]["group"]}

    def __enter__(self) -> "Client":
        return self

    def __exit__(self, exc_type: Type[Exception], exc_val: Exception, exc_tb: types.TracebackType) -> None:
        self._http_session.close()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['transmission_rpc']

package_data = \
{'': ['*']}

install_requires = \
['requests>=2.23.0,<3.0.0']

extras_require = \
{'docs': ['sphinx==6.1.3', 'sphinx-rtd-theme==1.2.0']}

setup_kwargs = {
    'name': 'transmission-rpc',
    'version': '4.1.0',
    'description': 'Python module that implements the Transmission bittorent client JSON-RPC protocol',
    'long_description': '# Transmission-rpc Readme\n\n[![PyPI](https://img.shields.io/pypi/v/transmission-rpc)](https://pypi.org/project/transmission-rpc/)\n[![Documentation Status](https://readthedocs.org/projects/transmission-rpc/badge/)](https://transmission-rpc.readthedocs.io/)\n[![ci](https://github.com/Trim21/transmission-rpc/workflows/ci/badge.svg)](https://github.com/Trim21/transmission-rpc/actions)\n[![PyPI - Python Version](https://img.shields.io/pypi/pyversions/transmission-rpc)](https://pypi.org/project/transmission-rpc/)\n[![Codecov branch](https://img.shields.io/codecov/c/github/Trim21/transmission-rpc/master)](https://codecov.io/gh/Trim21/transmission-rpc/branch/master)\n\n`transmission-rpc` is a python wrapper of on top of [transmission](https://github.com/transmission/transmission) JSON RPC protocol,\nhosted on GitHub at [github.com/trim21/transmission-rpc](https://github.com/trim21/transmission-rpc)\n\n## Introduction\n\n`transmission-rpc` is a python module implementing the json-rpc client protocol for the BitTorrent client Transmission.\n\nSupport 14 <= rpc version <= 17 (2.40 <= transmission version <= 4.0.1),\nshould works fine with newer rpc version but some new feature may be missing.\n\n## versioning\n\n`transmission-rpc` follow [Semantic Versioning](https://semver.org/),\nreport an issue if you found unexpected API break changes at same major version.\n\n## Install\n\n```console\npip install transmission-rpc -U\n```\n\n## Documents\n\n<https://transmission-rpc.readthedocs.io/>\n\n## Contributing\n\nAll kinds of PRs (docs, feature, bug fixes and eta...) are most welcome.\n\n## License\n\n`transmission-rpc` is licensed under the MIT license.\n',
    'author': 'Trim21',
    'author_email': 'i@trim21.me',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/Trim21/transmission-rpc',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



from typing import List, Literal, Optional

from transmission_rpc.types import Container


class Stats(Container):
    @property
    def uploaded_bytes(self) -> int:
        return self.fields["uploadedBytes"]

    @property
    def downloaded_bytes(self) -> int:
        return self.fields["downloadedBytes"]

    @property
    def files_added(self) -> int:
        return self.fields["filesAdded"]

    @property
    def session_count(self) -> int:
        return self.fields["sessionCount"]

    @property
    def seconds_active(self) -> int:
        return self.fields["secondsActive"]


class SessionStats(Container):
    # https://github.com/transmission/transmission/blob/main/docs/rpc-spec.md
    # 42-session-statistics

    @property
    def active_torrent_count(self) -> int:
        return self.fields["activeTorrentCount"]

    @property
    def download_speed(self) -> int:
        return self.fields["downloadSpeed"]

    @property
    def paused_torrent_count(self) -> int:
        return self.fields["pausedTorrentCount"]

    @property
    def torrent_count(self) -> int:
        return self.fields["torrentCount"]

    @property
    def upload_speed(self) -> int:
        return self.fields["uploadSpeed"]

    @property
    def cumulative_stats(self) -> Stats:
        return Stats(fields=self.fields["cumulative-stats"])

    @property
    def current_stats(self) -> Stats:
        return Stats(fields=self.fields["current-stats"])


class Units(Container):
    # 4 strings: KB/s, MB/s, GB/s, TB/s
    @property
    def speed_units(self) -> List[str]:
        return self.fields["speed-units"]

    # number of bytes in a KB (1000 for kB; 1024 for KiB)
    @property
    def speed_bytes(self) -> int:
        return self.fields["speed-bytes"]

    # 4 strings: KB/s, MB/s, GB/s, TB/s
    @property
    def size_units(self) -> List[str]:
        return self.fields["size-units"]

    # number of bytes in a KB (1000 for kB; 1024 for KiB)
    @property
    def size_bytes(self) -> int:
        return self.fields["size-bytes"]

    # 4 strings: KB/s, MB/s, GB/s, TB/s
    @property
    def memory_units(self) -> List[str]:
        return self.fields["memory-units"]

    # number of bytes in a KB (1000 for kB; 1024 for KiB)
    @property
    def memory_bytes(self) -> int:
        return self.fields["memory-bytes"]


class Session(Container):
    """
    Session is a class holding the session data for a Transmission daemon.

    Access the session field can be done through attributes.
    The attributes available are the same as the session arguments in the
    Transmission RPC specification, but with underscore instead of hyphen.


    get ``'download-dir'`` with ``session.download_dir``.

    .. code-block:: python

        session = Client().get_session()

        current = session.download_dir

    https://github.com/transmission/transmission/blob/main/docs/rpc-spec.md#41-session-arguments

    Warnings
    --------
    setter on session's properties has been removed, please use ``Client().set_session()`` instead
    """

    @property
    def alt_speed_down(self) -> int:
        """max global download speed (KBps)"""
        return self.fields["alt-speed-down"]

    @property
    def alt_speed_enabled(self) -> bool:
        # true means use the alt speeds
        return self.fields["alt-speed-enabled"]

    @property
    def alt_speed_time_begin(self) -> int:
        """when to turn on alt speeds (units: minutes after midnight)"""
        return self.fields["alt-speed-time-begin"]

    @property
    def alt_speed_time_day(self) -> int:
        """what day(s) to turn on alt speeds (look at tr_sched_day)"""
        return self.fields["alt-speed-time-day"]

    @property
    def alt_speed_time_enabled(self) -> bool:
        """true means the scheduled on/off times are used"""
        return self.fields["alt-speed-time-enabled"]

    @property
    def alt_speed_time_end(self) -> int:
        """when to turn off alt speeds (units: same)"""
        return self.fields["alt-speed-time-end"]

    @property
    def alt_speed_up(self) -> int:
        """max global upload speed (KBps)"""
        return self.fields["alt-speed-up"]

    @property
    def blocklist_enabled(self) -> bool:
        """true means enabled"""
        return self.fields["blocklist-enabled"]

    @property
    def blocklist_size(self) -> int:
        """int of rules in the blocklist"""
        return self.fields["blocklist-size"]

    @property
    def blocklist_url(self) -> str:
        """location of the blocklist to use for `blocklist-update`"""
        return self.fields["blocklist-url"]

    @property
    def cache_size_mb(self) -> int:
        """maximum size of the disk cache (MB)"""
        return self.fields["cache-size-mb"]

    @property
    def config_dir(self) -> str:
        """location of transmission's configuration directory"""
        return self.fields["config-dir"]

    @property
    def dht_enabled(self) -> bool:
        """true means allow dht in public torrents"""
        return self.fields["dht-enabled"]

    @property
    def download_dir(self) -> str:
        """default path to download torrents"""
        return self.fields["download-dir"]

    @property
    def download_dir_free_space(self) -> int:
        """**DEPRECATED** Use the `free-space` method instead."""
        return self.fields["download-dir-free-space"]

    @property
    def download_queue_enabled(self) -> bool:
        """if true, limit how many torrents can be downloaded at once"""
        return self.fields["download-queue-enabled"]

    @property
    def download_queue_size(self) -> int:
        """max int of torrents to download at once (see download-queue-enabled)"""
        return self.fields["download-queue-size"]

    @property
    def encryption(self) -> Literal["required", "preferred", "tolerated"]:
        return self.fields["encryption"]

    @property
    def idle_seeding_limit_enabled(self) -> bool:
        """true if the seeding inactivity limit is honored by default"""
        return self.fields["idle-seeding-limit-enabled"]

    @property
    def idle_seeding_limit(self) -> int:
        """torrents we're seeding will be stopped if they're idle for this long"""
        return self.fields["idle-seeding-limit"]

    @property
    def incomplete_dir_enabled(self) -> bool:
        """true means keep torrents in incomplete-dir until done"""
        return self.fields["incomplete-dir-enabled"]

    @property
    def incomplete_dir(self) -> str:
        """path for incomplete torrents, when enabled"""
        return self.fields["incomplete-dir"]

    @property
    def lpd_enabled(self) -> bool:
        """true means allow Local Peer Discovery in public torrents"""
        return self.fields["lpd-enabled"]

    @property
    def peer_limit_global(self) -> int:
        """maximum global int of peers"""
        return self.fields["peer-limit-global"]

    @property
    def peer_limit_per_torrent(self) -> int:
        """maximum global int of peers"""
        return self.fields["peer-limit-per-torrent"]

    @property
    def peer_port_random_on_start(self) -> bool:
        """true means pick a random peer port on launch"""
        return self.fields["peer-port-random-on-start"]

    @property
    def peer_port(self) -> int:
        """port int"""
        return self.fields["peer-port"]

    @property
    def pex_enabled(self) -> bool:
        """true means allow pex in public torrents"""
        return self.fields["pex-enabled"]

    @property
    def port_forwarding_enabled(self) -> bool:
        """true means ask upstream router to forward the configured peer port to transmission using UPnP or NAT-PMP"""
        return self.fields["port-forwarding-enabled"]

    @property
    def queue_stalled_enabled(self) -> bool:
        """whether or not to consider idle torrents as stalled"""
        return self.fields["queue-stalled-enabled"]

    @property
    def queue_stalled_minutes(self) -> int:
        """torrents that are idle for N minutes aren't counted toward seed-queue-size or download-queue-size"""
        return self.fields["queue-stalled-minutes"]

    @property
    def rename_partial_files(self) -> bool:
        """true means append `.part` to incomplete files"""
        return self.fields["rename-partial-files"]

    @property
    def rpc_version_minimum(self) -> int:
        """the minimum RPC API version supported"""
        return self.fields["rpc-version-minimum"]

    @property
    def rpc_version(self) -> int:
        """the current RPC API version"""
        return self.fields["rpc-version"]

    @property
    def script_torrent_done_enabled(self) -> bool:
        """whether or not to call the `done` script"""
        return self.fields["script-torrent-done-enabled"]

    @property
    def script_torrent_done_filename(self) -> str:
        """filename of the script to run"""
        return self.fields["script-torrent-done-filename"]

    @property
    def seed_queue_enabled(self) -> bool:
        """if true, limit how many torrents can be uploaded at once"""
        return self.fields["seed-queue-enabled"]

    @property
    def seed_queue_size(self) -> int:
        """max int of torrents to uploaded at once (see seed-queue-enabled)"""
        return self.fields["seed-queue-size"]

    @property
    def seedRatioLimit(self) -> float:
        """the default seed ratio for torrents to use"""
        return self.fields["seedRatioLimit"]

    @property
    def seedRatioLimited(self) -> bool:
        """true if seedRatioLimit is honored by default"""
        return self.fields["seedRatioLimited"]

    @property
    def speed_limit_down_enabled(self) -> bool:
        """true means enabled"""
        return self.fields["speed-limit-down-enabled"]

    @property
    def speed_limit_down(self) -> int:
        """max global download speed (KBps)"""
        return self.fields["speed-limit-down"]

    @property
    def speed_limit_up_enabled(self) -> bool:
        """true means enabled"""
        return self.fields["speed-limit-up-enabled"]

    @property
    def speed_limit_up(self) -> int:
        """max global upload speed (KBps)"""
        return self.fields["speed-limit-up"]

    @property
    def start_added_torrents(self) -> bool:
        """true means added torrents will be started right away"""
        return self.fields["start-added-torrents"]

    @property
    def trash_original_torrent_files(self) -> bool:
        """true means the .torrent file of added torrents will be deleted"""
        return self.fields["trash-original-torrent-files"]

    # see below
    @property
    def units(self) -> Units:
        return self.fields["units"]

    @property
    def utp_enabled(self) -> bool:
        """true means allow utp"""
        return self.fields["utp-enabled"]

    @property
    def version(self) -> str:
        """long version str `$version ($revision)`"""
        return self.fields["version"]

    @property
    def default_trackers(self) -> Optional[list]:
        """
        list of default trackers to use on public torrents
        new at rpc-version 17
        """
        return self.get("default-trackers")

    @property
    def rpc_version_semver(self) -> Optional[str]:
        """
        the current RPC API version in a semver-compatible str
        new at rpc-version 17
        """
        return self.get("rpc-version-semver")

    @property
    def script_torrent_added_enabled(self) -> Optional[bool]:
        """
        whether or not to call the `added` script
        new at rpc-version 17
        """
        return self.get("script-torrent-added-enabled")

    @property
    def script_torrent_added_filename(self) -> Optional[str]:
        """
        filename of the script to run
        new at rpc-version 17
        """
        return self.get("script-torrent-added-filename")

    @property
    def script_torrent_done_seeding_enabled(self) -> Optional[bool]:
        """
        whether or not to call the `seeding-done` script
        new at rpc-version 17
        """
        return self.get("script-torrent-done-seeding-enabled")

    @property
    def script_torrent_done_seeding_filename(self) -> Optional[str]:
        """
        filename of the script to run
        new at rpc-version 17
        """
        return self.get("script-torrent-done-seeding-filename")
